#!/usr/bin/env python
import cStringIO
import math
import new
import os
import signal
import sys
import Tkinter
import tkFileDialog
import tkMessageBox
import traceback

import Image
import ImageTk
import mapnik

from pysmap.common import *
from pysmap import gisbrowse
from pysmap import settings
from pysmap import tracks

mapFile = settings.get("mapnikDefs")

tileSize = 1024
gutterSize = 200

tileCachePath = os.path.join(os.environ["HOME"], ".tiles")


class Deferred(object):
	"""is a sentinel for pending tiles.
	"""
	def __init__(self, tilex, tiley):
		self.tilex, self.tiley = tilex, tiley


def makeTile(*imArgs, **imKw):
	"""returns a PIL image for the constructor arguments given.
	"""
	res = Image.open(*imArgs, **imKw)
	res = res.crop((gutterSize, gutterSize,
		tileSize+gutterSize, tileSize+gutterSize))
	return res


class TileSet(object):
	"""is a set of tiles of common scale.

	map and projection are mapnik objects.
	"""
	def __init__(self, map, projection, metersPerPixel=5):
		self.map, self.proj = map, projection
		self.metersPerPixel = float(metersPerPixel)
		self.cacheBase = os.path.join(tileCachePath, "%.2f"%(metersPerPixel))
		self.pendingComputations, self.finishedQueue = {}, []
		if not os.path.exists(self.cacheBase):
			os.makedirs(self.cacheBase)
		self.tileCache = {}
		signal.signal(signal.SIGCHLD, self.handleFinishedChild)

	def getBboxForTile(self, longIndex, latIndex):
		x0, y0 = longIndex*tileSize, -latIndex*tileSize
		return mapnik.Envelope(
			(x0-gutterSize)*self.metersPerPixel, 
			(y0+gutterSize)*self.metersPerPixel,
			(x0+tileSize+gutterSize)*self.metersPerPixel, 
			(y0-tileSize-gutterSize)*self.metersPerPixel)

	def getCanvasCForGeo(self, long, lat):
		pc = self.proj.forward(mapnik.Coord(long, lat))
		return pc.x/self.metersPerPixel, -pc.y/self.metersPerPixel

	def getGeoCForCanvasC(self, x, y):
		pc = self.proj.inverse(mapnik.Coord(x*self.metersPerPixel, 
			y*self.metersPerPixel))
		return pc.x, -pc.y

	def handleFinishedChild(self, sigNum, stackFrame):
		"""is a signal handler for SIGCHLD.
		"""
		try:
			while 1:
				pid, status = os.waitpid(-1, os.WNOHANG)
				if pid==0:
					break
				self.finishedQueue.append(self.pendingComputations.pop(pid))
		except os.error: # no child processes, we're done anyway
			pass

	def allDone(self):
		return not self.pendingComputations

	def _computeTile(self, tilex, tiley):
		bbox = self.getBboxForTile(tilex, tiley)
		self.map.zoom_to_box(bbox)
		im = mapnik.Image(tileSize+2*gutterSize, tileSize+2*gutterSize)
		mapnik.render(self.map, im)
# No idea what the following means -- the arguments don't seem to matter.
		view = im.view(0, 0, tileSize, tileSize)
		res = makeTile(cStringIO.StringIO(view.tostring('png')))
		return res

	def loadTile(self, tilex, tiley):
		cachePath = os.path.join(self.cacheBase, "%d-%d.png"%(tilex, tiley))
		if os.path.exists(cachePath):
			self.tileCache[tilex, tiley] = Image.open(cachePath)
		else:
			chpid = os.fork()
			if chpid: # parent
				self.pendingComputations[chpid] = (tilex, tiley)
			else: # child
				try:
					data = self._computeTile(tilex, tiley)
					data.save(cachePath, format="png")
				finally:
					os._exit(0)
			self.tileCache[tilex, tiley] = Deferred(tilex, tiley)
		
	def getTileForIndex(self, tilex, tiley):
		if (tilex, tiley) in self.tileCache:
			content = self.tileCache[tilex, tiley]
			return content
		else:
			self.loadTile(tilex, tiley)
			return self.tileCache[tilex, tiley]


class TiledMap(Tkinter.Frame):
	def __init__(self, master):
		self.lastCenter = None
		Tkinter.Frame.__init__(self, master)
		self.parent = master
		self.zoomListeners = []
		self._buildWidgets()
		self.afterId = self.after(100, self.checkForTiles)

	def _buildWidgets(self):
		self.piczone = Tkinter.Canvas(self)
		self.setTileSet(None)
		self.piczone.config(xscrollincrement=1, yscrollincrement=1)
		self.piczone.config(background="#FFFFFF")
		self.piczone.bind('<Button-3>', self._startmove)
		self.piczone.bind('<B3-Motion>', self._domove)
		self.piczone.bind('<ButtonRelease-3>', self._endmove)
#		self.piczone.bind('<Motion>', self._displayPos)
		self.piczone.pack(expand=1, fill=Tkinter.BOTH)

	def addZoomListener(self, listener):
		"""registers a function to call when the mapping from geo coordinates
		to canvas coordinates changes.

		The listener is called with the new tileSet.
		"""
		self.zoomListeners.append(listener)

	def _zoomChanged(self):
		for f in self.zoomListeners:
			f(self.tileSet)

	def getCenterGeo(self):
		return self.tileSet.getGeoCForCanvasC(
			self.piczone.canvasx(self.piczone.winfo_width()/2), 
			self.piczone.canvasy(self.piczone.winfo_height()/2))

	def setTileSet(self, tileSet):
		if self.lastCenter:
			self.lastCenter = self.getCenterGeo()
		self.tileSet, self.tiles = tileSet, {}
		self.piczone.delete(Tkinter.ALL)
		if self.lastCenter:
			self.goTo(*self.lastCenter)
		self._zoomChanged()
		
	def _displayPos(self, event):
		print self.piczone.canvasx(event.x), self.piczone.canvasy(event.y)
		
	def _startmove(self, event):
		self._lastx = event.x
		self._lasty = event.y

	def _domove(self, event):
		try:
			dx = event.x-self._lastx
			dy = event.y-self._lasty
			self.piczone.xview(Tkinter.SCROLL, -dx, Tkinter.UNITS)
			self.piczone.yview(Tkinter.SCROLL, -dy, Tkinter.UNITS)
			self._lastx = event.x
			self._lasty = event.y
		except AttributeError:
			pass

	def _endmove(self, event):
		try:
			dx = event.x-self._lastx
			dy = event.y-self._lasty
			self.piczone.xview(Tkinter.SCROLL, -dx, Tkinter.UNITS)
			self.piczone.yview(Tkinter.SCROLL, -dy, Tkinter.UNITS)
			del(self._lastx)
			del(self._lasty)
			self.updateTiles()
		except:
			pass

	def setCanvasPosition(self, x0, y0):
# does it really need to be that hard?
		self.piczone.xview(Tkinter.SCROLL, int(-self.piczone.canvasx(0)), 
			Tkinter.UNITS)
		self.piczone.xview(Tkinter.SCROLL, int(x0)-self.winfo_width()/2, 
			Tkinter.UNITS)
		self.piczone.yview(Tkinter.SCROLL, int(-self.piczone.canvasy(0)), 
			Tkinter.UNITS)
		self.piczone.yview(Tkinter.SCROLL, int(y0)-self.winfo_height()/2, 
			Tkinter.UNITS)

	def _displayImage(self, tilex, tiley, data):
		self.tiles[tilex, tiley] = ImageTk.PhotoImage(data)
		self.piczone.lower(
			self.piczone.create_image(tilex*tileSize, tiley*tileSize, anchor="nw",
			image=self.tiles[tilex, tiley]))

	def checkForTiles(self):
		if self.tileSet is None:
			return
		try:
			if self.tileSet.allDone():
				self.parent.endActivity()
			if self.tileSet.finishedQueue:
				self.config(cursor="watch")
				while self.tileSet.finishedQueue:
					tilex, tiley = self.tileSet.finishedQueue.pop()
					self.tileSet.loadTile(tilex, tiley)
					data = self.tileSet.getTileForIndex(tilex, tiley)
					self._displayImage(tilex, tiley, data)
				self.config(cursor="right_ptr")
		finally:
			self.after(100, self.checkForTiles)

	def showImage(self, tilex, tiley):
		if self.tileSet is not None and not (tilex, tiley) in self.tiles:
			data = self.tileSet.getTileForIndex(tilex, tiley)
			if not isinstance(data, Deferred):
				self._displayImage(tilex, tiley, data)

	def goTo(self, long, lat):
		if self.tileSet is None:
			return
		try:
			self.lastCenter = (long, lat)
			self.setCanvasPosition(*self.tileSet.getCanvasCForGeo(long, lat))
			self.config(cursor="watch")
			try:
				self.updateTiles()
			finally:
				self.config(cursor="left_ptr")
		except:
			traceback.print_exc()

	def updateTiles(self):
		if self.tileSet is None:
			return
		self.parent.startActivity()
		xLo, xHi = int(math.floor(self.piczone.canvasx(0)/tileSize)),\
		 int(math.floor(self.piczone.canvasx(self.winfo_width()))/tileSize)
		yLo, yHi = int(math.floor(self.piczone.canvasy(0))/tileSize),\
		 int(math.floor(self.piczone.canvasy(self.winfo_height())/tileSize))
		for x in range(xLo, xHi+1):
			for y in range(yLo-1, yHi+1):
				self.showImage(x, y)


class TiledMapWithTrack(TiledMap):
	def __init__(self, master):
		self.track = None
		self.trackItems = {}
		TiledMap.__init__(self, master)
		self.addZoomListener(self._updateTrack)
		self.piczone.bind('<Button-1>', self._startSelect)
		self.piczone.bind('<B1-Motion>', self._moveSelect)
		self.piczone.bind('<ButtonRelease-1>', self._endSelect)

	def hasTrack(self):
		return self.track is not None
	
	def setTrack(self, track):
		self.track = track
		self._updateTrack()

	def gotoTrackStart(self):
		_, long, lat, _ = self.track.track[0]
		self.goTo(long, lat)

	def _updateTrack(self, ignored=None):
		self.piczone.delete("TRACKPOINT")
		self.piczone.delete("ACTIVE_TRACKPOINT")
		if self.track is None:
			return
		self.trackItems = {}
		for index, (stamp, long, lat, _) in enumerate(self.track):
			x,y = self.tileSet.getCanvasCForGeo(long, lat)
			self.trackItems[self.piczone.create_rectangle(x-2, y-2, x+2, y+2,
				fill="green", tags=("TRACKPOINT",))] = index

	def _startSelect(self, event):
		self._selectX0 = self.piczone.canvasx(event.x)
		self._selectY0 = self.piczone.canvasy(event.y)

	def _moveSelect(self, event):
		self.piczone.delete("SELRECT")
		curX0 = self.piczone.canvasx(event.x)
		curY0 = self.piczone.canvasy(event.y)
		self.piczone.create_rectangle(self._selectX0, self._selectY0, curX0,
			curY0, tags=("SELRECT",))

	def _endSelect(self, event):
		if not hasattr(self, "_selectX0"): # stray up event
			return
		curX0 = self.piczone.canvasx(event.x)
		curY0 = self.piczone.canvasy(event.y)
		rect = self._selectX0, self._selectY0, curX0, curY0
		del self._selectX0
		del self._selectY0
		self.piczone.delete("SELRECT")
		if event.state&4:  # Control pressed
			self.unselectTrackpoints(rect)
		else:
			self.selectTrackpoints(rect)

	def getTrackpointsWithin(self, rect=None):
		"""returns a set of canvas item indices for all trackpoints within rect.

		rect is specified as (x0, y0, x1, y1).  It can also be None to
		get all trackpoints.
		"""
		res = set(self.trackItems)
		if rect is not None:
			itemsWithin = set(self.piczone.find_enclosed(*rect))
			res = itemsWithin&res
		return res

	def selectTrackpoints(self, rect=None):
		for id in self.getTrackpointsWithin(rect):
			self.piczone.itemconfig(id, tags=("ACTIVE_TRACKPOINT",))
		self.piczone.itemconfig('ACTIVE_TRACKPOINT', fill="red")

	def unselectTrackpoints(self, rect=None):
		for id in self.getTrackpointsWithin(rect):
			self.piczone.itemconfig(id, tags=("TRACKPOINT",))
		self.piczone.itemconfig('TRACKPOINT', fill="green")

	def _deleteTPs(self, tpInds):
		toDelete = set(self.trackItems[ind] for ind in tpInds)
		self.track.deleteByIndices(toDelete)
		self._updateTrack()

	def deleteSelectedTPs(self):
		self._deleteTPs(self.piczone.find_withtag("ACTIVE_TRACKPOINT"))

	def deleteUnselectedTPs(self):
		self._deleteTPs(self.piczone.find_withtag("TRACKPOINT"))


class UI(Tkinter.Tk):
	def __init__(self, osmMap, initPos):
		Tkinter.Tk.__init__(self)
		self.osmMap = osmMap
		self.blinkTimer = None
		self._buildVariables()
		self._buildWidgets()
		self._doBindings()
		self._updateScale()
		self._makeMenu()
		self.bind("<Map>", lambda ev: self.oMap.goTo(*initPos))

	def _buildVariables(self):
		self.saveCloaked = Tkinter.IntVar()
		self.saveCloaked.set(settings.get("state", "saveCloaked"))

	def _makeMenu(self):
		menubar = Tkinter.Menu(self)

		filemenu = Tkinter.Menu(menubar, tearoff=0)
		filemenu.add_command(label="Read GPX...", command=self._loadTrack,
			accelerator="^O")
		filemenu.add_command(label="Save GPX...", command=self._saveTrack,
			accelerator="^S")
		filemenu.add_checkbutton(label="...fuzz times", variable=self.saveCloaked)
		filemenu.add_command(label="Quit", command=self.quit,
			accelerator="^Q")
		menubar.add_cascade(label="File", menu=filemenu)

		editmenu = Tkinter.Menu(menubar, tearoff=0)
		editmenu.add_command(label="Delete selected", command=self._makeCallback(
			self.oMap.deleteSelectedTPs), accelerator="^X")
		editmenu.add_command(label="Delete unselected", command=self._makeCallback(
			self.oMap.deleteUnselectedTPs), accelerator="^Y")
		menubar.add_cascade(label="Edit", menu=editmenu)

		gomenu = Tkinter.Menu(menubar, tearoff=0)
		gomenu.add_command(label="Track start", command=self._gotoTrackStart,
			accelerator="Alt-T")
		menubar.add_cascade(label="Go", menu=gomenu)

		self.config(menu=menubar)

	def _makeLocator(self):
		butframe = Tkinter.Frame(self)
		self.renderLabel = Tkinter.Label(butframe, text="C", background="black")
		self.renderLabel.pack(side=Tkinter.LEFT)
		self.placeEntry = Tkinter.Entry(butframe)
		self.placeEntry.insert(Tkinter.END, "Dossenheim")
		self.placeEntry.pack(expand=True, fill=Tkinter.X, side=Tkinter.LEFT)
		self.placeEntry.focus()
		searchButton = Tkinter.Button(butframe, command=self._doSearch,
			text="...as Settlement")
		searchButton.pack(side=Tkinter.LEFT)
		return butframe
	
	def _makeScaler(self):
		frame = Tkinter.Frame(self)
		self.curScale = Tkinter.IntVar()
		self.curScale.set(settings.get("state", "initalScale"))
		for scale in settings.definedScales:
			Tkinter.Radiobutton(frame, text="%d m/pixel"%scale,
				variable=self.curScale, value=scale, command=self._updateScale
				).pack(anchor=Tkinter.W)
		return frame

	def _buildWidgets(self):
		self.oMap = TiledMapWithTrack(self)
		self.oMap.grid(row=2, column=2, sticky=Tkinter.NW+Tkinter.SE)
		self.rowconfigure(2, weight=1)
		self.columnconfigure(2, weight=1)
		self._makeLocator().grid(row=3, column=2, sticky=Tkinter.E+Tkinter.W)
		self._makeScaler().grid(row=2, column=3, sticky=Tkinter.S+Tkinter.N)

	def _makeCallback(self, fct, args=()):
		def callback(ev=None):
			self.config(cursor="watch")
			try:
				return fct(*args)
			finally:
				self.config(cursor="left_ptr")
		return callback

	def _doBindings(self):
		self.bind("<Control-q>", lambda ev: self.quit())
		self.bind("<Control-o>", lambda ev: self._loadTrack())
		self.bind("<Control-s>", self._makeCallback(self._saveTrack))
		self.bind("<Control-x>", self._makeCallback(self.oMap.deleteSelectedTPs))
		self.bind("<Control-y>", self._makeCallback(self.oMap.deleteUnselectedTPs))
		self.bind("<Alt-T>", lambda ev: self._gotoTrackStart)
		self.placeEntry.bind("<Return>", lambda ev: self._doSearch())
		self.focus()

	def _updateScale(self):
		self.oMap.setTileSet(TileSet(self.osmMap, mapnikProjection, 
			self.curScale.get()))

	def _doSearch(self):
		self.config(cursor="watch")
		self.update_idletasks()
		try:
			term = self.placeEntry.get().strip()
			if term:
				term = term+"%"
				context = gisbrowse.GisContext(gisbrowse.gisDSN, mapnikProjection)
				res = gisbrowse.Settlements(context, term)
				if len(res)==0:
					tkMessageBox.showerror("Not Found", "No match for %s"%repr(
						term))
				elif len(res)==1:
					self.oMap.goTo(*res[0][1:])
				else:
					dia = gisbrowse.GeoBrowser(self, res)
					if dia.placeName:
						self.oMap.goTo(*dia.geoCoos)
		finally:
			self.config(cursor="left_ptr")

	def _saveTrack(self):
		if not self.oMap.hasTrack():
			return tkMessageBox.showerror("No Track", "Load a track before"
				" trying to save it.")
		fName = tkFileDialog.asksaveasfilename(parent=self, title="Save GPX",
			filetypes=[("GPX files", "*.gpx")])
		if not fName:
			return
		try:
			self.oMap.track.save(fName, self.saveCloaked.get())
		except ReportableError, msg:
			tkMessageBox.showerror("Error while saving", 
				unicode(msg))

	def _loadTrack(self):
		fName = tkFileDialog.askopenfilename(parent=self, title="Load GPX",
			filetypes=[("GPX files", "*.gpx")])
		if not fName:
			return
		return self.loadTrackFromFile(fName)

	def loadTrackFromFile(self, fName):
		self.config(cursor="watch")
		self.oMap.config(cursor="watch")
		self.update_idletasks()
		try:
			try:
				self.oMap.setTrack(tracks.makeGPXFromXML(fName))
			except Exception, msg:
				tkMessageBox.showerror("GPX load failed",
					"An error occurred while trying to load %s:\n%s"%(
						fName, str(msg)))
		finally:
			self.config(cursor="left_ptr")
			self.oMap.config(cursor="left_ptr")

	def _gotoTrackStart(self):
		if self.oMap.hasTrack():
			try:
				self.oMap.gotoTrackStart()
			except Exception:
				self.tkMessageBox.showerror("Track bad",
					"Empty or bad track")
		else:
			self.tkMessageBox.showerror("Load track first",
				"Load a track before going to its start")

	def startActivity(self):
		if self.blinkTimer is not None:
			return
		colorSequence = {
			"black": "#330000", "#330000": "#550000", "#550000": "#770000",
			"#770000": "#990000", "#990000": "#AA0000", "#AA0000": "#BB0000",
			"#BB0000": "#CC0000", "#CC0000": "#DD0000", "#DD0000": "#EE0000",
			"#EE0000": "red", "red": "#EE0001", "#EE0001": "#DD0001", "#DD0001":
			"#CC0001", "#CC0001": "#BB0001", "#BB0001": "#AA0001", "#AA0001":
			"#990001", "#990001": "#770001", "#770001": "#550001", "#550001":
			"#330001", "#330001": "black",}

		def doBlink():
			try:
				self.renderLabel.config(background=colorSequence.get(
					self.renderLabel.cget("background"), "red"))
			finally:
				self.blinkTimer = self.after(100, doBlink)
		doBlink()
	
	def endActivity(self):
		if self.blinkTimer is not None:
			self.after_cancel(self.blinkTimer)
			self.renderLabel.config(background="black")
			self.blinkTimer = None


mapnikProjection = mapnik.Projection("+proj=merc +a=6378137 +b=6378137"
	" +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m"
	" +nadgrids=@null +no_defs +over")


def runUI(opts, args):
	osmMap = mapnik.Map(tileSize+2*gutterSize, tileSize+2*gutterSize)
	mapnik.load_map(osmMap, settings.get("mapnikDefs").encode("utf-8"))
	panner = UI(osmMap, settings.get("state", "initialPos"))
	panner.geometry(settings.get("state", "geometry"))
	panner.title("Simple Map")
	panner.loadTrackFromFile(args[0])
	panner.mainloop()
	settings.set("state", "saveCloaked", (panner.saveCloaked.get() and 'True')
		or 'False')
	settings.set("state", "geometry", panner.geometry())
	settings.set("state", "initalScale", str(panner.curScale.get()))
	if panner.oMap.lastCenter:
		settings.set("state", "initialPos", 
			"%.9f,%.9f"%panner.oMap.getCenterGeo())
	settings.save()


def printConfigDocs():
	"""outputs the configuration documentation to stdout.
	"""
	import fancyconfig
	print fancyconfig.makeTxtDocs(settings._config)


def parseCommandLine():
	from optparse import OptionParser
	parser = OptionParser(usage="%%prog {<option>} [<gpx-name>]"
		" -- a GUI for offline mapnik rendering.")
	parser.add_option("-H", "--help-config", help="Show help on items"
		" available for %s and exit"%settings.userConfigName,
		dest="showConfig", action="store_true", default=False)
	opts, args = parser.parse_args()
	if opts.showConfig:
		printConfigDocs()
		sys.exit(0)
	if len(args)>1:
		parser.print_help()
		sys.exit(1)
	return opts, args


def main():
	opts, args = parseCommandLine()
	runUI(opts, args)


if __name__=="__main__":
	main()
