"""
Configuration file management and definition, common stuff.
"""


import os

from pysmap import fancyconfig
from pysmap.fancyconfig import (Configuration, DefaultSection, Section,
	StringConfigItem, ConfigItem, ParseError, IntConfigItem, FloatConfigItem,
	BooleanConfigItem)


class Error(Exception):
	"""is the base class for all of our exceptions.
	"""


userConfigName = os.path.join(os.environ.get("HOME", "/"), ".pysmap")

definedScales = [1,5,10,20,50,100,500]

class CoordinateConfigItem(ConfigItem):
	typedesc = "coordinates (long,lat)"

	def _parse(self, value):
		try:
			long, lat = map(float, value.split(","))
		except ValueError:
			raise ParseError("%s is not a valid coordinate specification"%
				value)
		return long, lat
	
	def _unparse(self, value):
		return "%.9f,%.9f"%value


class EnumeratedIntConfigItem(IntConfigItem):
	def __init__(self, name, **kwargs):
		if "options" in kwargs:
			self.options = set(kwargs.pop("options"))
		IntConfigItem.__init__(self, name, **kwargs)
	
	def _parse(self, value):
		value = IntConfigItem._parse(self, value)
		if not value in self.options:
			raise ParseError("%s is not in the list of allowed values, %s"%(
				value, self.options))
		return value

_config = Configuration(
	DefaultSection("Everything that doesn't fit anywhere else.",
		StringConfigItem("mapnikDefs", "", description="Path to"
			" the mapnik render definition."),
	),
	Section("db", "Settings defining database access",
		StringConfigItem("dsn", default="dbname=gis", description=
			"String defining the connection to the GIS database for search"
			" operations.  The format is <key>=<value>{,<key>=<value>},"
			" where keys include dbname, host, port, user, password and more."
			" See psycopg documentation for details."),
	),
	Section("export", "Settings related to exporting tracks",
		IntConfigItem("startjitter", default="30", description=
			"Width in days of the interval in which the start point of a track"
			" will be shifted on time-cloaked gpx export."),
		FloatConfigItem("pointInterval", default="10", description=
			"Number of seconds to track/waypoints will be separated by in"
			" time-cloaked export"),
	),
	Section("state", "Application state (these will be overwritten on"
		" exit)",
		StringConfigItem("geometry", "900x550", 
			description="Window geometry"),
		CoordinateConfigItem("initialPos", "10.0000,50.0000",
			description="Initial coordinates (long, lat)"),
		EnumeratedIntConfigItem("initalScale", default="5",
			options=definedScales, description="Initial scale of map"),
		BooleanConfigItem("saveCloaked", default="True",
			description="Save GPX files with 'normalized' timestamps?"),
	),
)

def save():
	_config.saveUserConfig(userConfigName)

set = _config.set
get = _config.get
getitem = _config.getitem


if __name__=="__main__":
	print fancyconfig.makeTxtDocs(_config)
else:
	fancyconfig.readConfiguration(_config, None, userConfigName)


