"""
Support for tracks (in gpx format, for now).
"""

from xml import sax
from xml.sax.handler import ContentHandler
import datetime
import itertools
import random
import sys
import time
import traceback

from pysmap import settings
from pysmap.common import *


try:
	try:
		from xml.etree import ElementTree
	except ImportError:
		try:
			import cElementTree as ElementTree
		except ImportError:
			from elementtree import ElementTree
	E = ElementTree.SubElement
except ImportError:
	E = None


class NamedNode:
	"""is a helper class for NodeBuilder to change node names from
	handling functions.
	"""
	def __init__(self, name, node):
		self.name, self.node = name, node


class NodeBuilder(ContentHandler):
	"""a node builder is a content handler working more along the
	lines of conventional parse tree builders.

	This means that for every element we want handled, there is a
	method _make_<elementname>(name, attrs, children) that receives
	the name of the element (so you can reuse implementations for
	elements behaving analogously), the attributes of the element
	and a list of children.  The children come in a list of tuples
	(name, content), where name is the element name and content
	is whatever the _make_x method returned for that element.
	Text nodes have a name of None.

	NodeBuilders support a limited id/idref mechanism.  Nodes with
	id will get entered in a dictionary and can be retrieved (as
	name/node pairs) via getById.  However, this only is possible
	after the element with the id has been closed.	There is no
	forward declaration.

	In general, text children are deleted when they are whitespace
	only, and they are joined to form a single one.  However, you
	can define a set keepWhitespaceNames containing the names of
	elements for which this is not wanted.

	In some cases, you want parents provide information to their
	children while they are constructed.  This is a bit clumsy, but
	for such cases, you can define a _start_<element> method that can
	leave something in a dictionary through the pushProperty method
	that can be retrieved by children through the getProperty method.
	When constructing the parent node, you must call popProperty
	on this.

	As an added hack, you can register nodes for
	addition to the nearest enclosing element of a type via
	registerDelayedChildren.  This is provided to allow methods
	change the tree higher up if necessary; Here's an example:
	You have <foo><bar><baz/></bar></foo>, and the handler for
	baz decides it wants to have a Bla sibling.  It can then call
	registerDelayedChild("bar", Bla()).

	On errors during node construction, the class will call a
	handleError method with a sys.exc_info tuple.
	"""
	keepWhitespaceNames = set()

	def __init__(self):
		ContentHandler.__init__(self)
		self.elementStack = []
		self.childStack = [[]]
		self.delayedChildren = {}
		self.properties = {}
		self.locator = None
		self.elementsById = {}

	def registerDelayedChild(self, parentName, child, atfront=False):
		"""adds child for addition to the next enclosing parentName element.
		"""
		if not self.delayedChildren.has_key(parentName):
			self.delayedChildren[parentName] = ([], [])
		if atfront:
			self.delayedChildren[parentName][0].append(child)
		else:
			self.delayedChildren[parentName][1].append(child)

	def pushProperty(self, propName, value):
		"""makes value available to node constructors under the name propName.

		It is recommended to use <element>.<name> as propname.
		"""
		self.properties.setdefault(propName, []).append(value)
	
	def popProperty(self, propName):
		"""retrieves (and removes) the last value pushed as propName.
		"""
		return self.properties[propName].pop()
	
	def getProperty(self, propName):
		"""returns the current value of the property propName.

		Non-existing properties will be signalled by raising an IndexError.
		"""
		try:
			return self.properties[propName][-1]
		except (IndexError, KeyError):
			raise IndexError("Property %s is not set"%propName)
	
	def handleError(self, exc_info):
		msg = ("Error while parsing XML at"
			" %d:%d (%s)"%(self.locator.getLineNumber(), 
				self.locator.getColumnNumber(), exc_info[1]))
		traceback.print_exc()
		raise Error(msg)
	
	def setDocumentLocator(self, locator):
		self.locator = locator

	def startElement(self, name, attrs):
		self.elementStack.append((name, attrs))
		self.childStack.append([])
		if hasattr(self, "_start_"+name):
			getattr(self, "_start_"+name)(name, attrs)

	def _enterNewNode(self, name, attrs, newNode):
			if isinstance(newNode, NamedNode):
				newChild = (newNode.name, newNode.node)
			else:
				newChild = (name, newNode)
			self.childStack[-1].append(newChild)
			if attrs.has_key("id"):
				if attrs["id"] in self.elementsById:
					raise Error("Duplicate id: %s"%attrs["id"])
				self.elementsById[attrs["id"]] = newChild

	def endElement(self, name):
		_, attrs = self.elementStack.pop()
		try:
			children = self.childStack.pop()
			if self.delayedChildren.has_key(name):
				children[:0] = self.delayedChildren[name][0]
				children.extend(self.delayedChildren[name][1])
				del self.delayedChildren[name]
			if name not in self.keepWhitespaceNames:
				children = self._cleanTextNodes(children)
			if hasattr(self, "_make_"+name):
				handler = getattr(self, "_make_"+name)
			elif hasattr(self, "_make_default"):
				handler = getattr(self, "_make_default")
			else:
				raise Error("No handler for %s"%name)
			newNode = handler(name, attrs, children)
			if newNode is not None:
				self._enterNewNode(name, attrs, newNode)
		except:
			self.handleError(sys.exc_info())

	def characters(self, content):
		self.childStack[-1].append((None, content))

	def getById(self, id):
		return self.elementsById[id]

	def getResult(self):
		return self.childStack[0][0][1]
	
	def getNodesDict(self, children):
		"""returns children as a dictionary of lists.

		children is a list of the type passed to the _make_xxx methods.
		"""
		res = {}
		for name, val in children:
			res.setdefault(name, []).append(val)

	def getContentWS(self, children):
		"""returns the entire text content of the node in a string without doing
		whitespace normalization.
		"""
		return "".join([n[1] for n in children if n[0] is None])

	def getContent(self, children):
		"""returns the entire text content of the node in a string.

		This probably won't do what you want in mixed-content models.
		"""
		return self.getContentWS(children).strip()

	def getWaitingChild(self, nodeName, maxLevels=100, startLevel=-1):
		"""returns the first child with nodeName waiting to be adopted in the
		childStack.

		maxLevels gives a maximum number of childStack levels we descend
		before giving up.

		If no matching child can be found, a NoWaitingChild exception is raised.
		"""
		if maxLevels==0:
			raise NoWaitingChild(nodeName)
		for name, element in self.childStack[startLevel]:
			if name==nodeName:
				return element
		return getWaitingChild(nodeName, maxLevels-1, startLevel-1)

	def _cleanTextNodes(self, children):
		"""joins adjacent text nodes and prunes whitespace-only nodes.
		"""
		cleanedChildren, text = [], []
		for type, node in children:
			if type is None:
				text.append(node)
			else:
				if text:
					chars = "".join(text)
					if chars.strip():
						cleanedChildren.append((None, chars))
					text = []
				cleanedChildren.append((type, node))
		chars = "".join(text)
		if chars.strip():
			cleanedChildren.append((None, chars))
		return cleanedChildren

	def filterChildren(self, children, targetNode):
		"""returns a list of children that are of type targetNode, and a
		list of all other children.
		"""
		return [child for child in children if child[0]==targetNode
			], [child for child in children if child[0]!=targetNode]

	def _processChildren(self, parent, name, childMap, children, 
			ignoreUnknownElements=False):
		"""adds children to parent.

		Parent is some class (usually a record.Record instance),
		childMap maps child names to methods to call the children with,
		and children is a sequence as passed to the _make_xxx methods.

		The function returns parent for convenience.
		"""
		for childName, val in children:
			try:
				childMap[childName](val)
			except KeyError:
				traceback.print_exc()
				if not ignoreUnknownElements:
					raise Error("%s elements may not have %s children"%(
						name, childName))
		return parent


class PropertyMixin(object):
	__properties = None

	def setProperty(self, name, value):
		if self.__properties is None:
			self.__properties = {}
		self.__properties[name] = value
	
	def getProperty(self, name):
		if self.__properties is None:
			self.__properties = {}
		return self.__properties.get(name)

	def hasProperty(self, name):
		if self.__properties is None:
			self.__properties = {}
		return name in self.__properties


class Track(PropertyMixin):
	"""is a track, including a name and a list of points.

	Points consist of tuples (time, long, lat), where time is a 
	unix timestamp.
	"""
	def __init__(self):
		self.name = ""
		self.points = []

	def __iter__(self):
		for p in self.points:
			yield p

	def add(self, tpt):
		"""adds a trackpoint.  
		
		A trackpoint is a tuple (time, long, lat [, ele])
		"""
		self.points.append(tpt)

	def addPoints(self, tptList):
		for tpt in tptList:
			self.add(tpt)


class GPX(PropertyMixin):
	"""models the relevant content of a GPX file.

	All tracks and track segments will be joined into one track.
	This track includes waypoints as well.  Tell them from trackpoints
	by the fact that it has a name in its properties.
	"""
	def __init__(self):
		self.track = []
		self.nextTime = None

	def __iter__(self):
		return iter(self.track)

	def addTrack(self, track):
		self.track.extend(track)

	def addTrackpoint(self, point):
		self.track.append(point)

	def deleteByIndices(self, indexSet):
		self.track = [p for ind, p in enumerate(self.track)
			if ind not in indexSet]

	def normalize(self):
		self.track.sort()

	def _setupForcedTime(self):
		"""configures the object for forged timestamps on export.

		To undo, simply set nextTime to None.
		"""
		if self.track:
			self.nextTime = self.track[0][0]+int((random.random()-0.5
				)*settings.get("export", "startjitter"))*86400
			self.pointInterval = settings.get("export", "pointInterval")

	def _iterTimes(self):
		if self.nextTime:
			while True:
				yield self.nextTime
				self.nextTime += self.pointInterval
		else:
			while True:
				yield None

	def _makeWaypoint(self, parent, wptObject, forceTime):
		t, long, lat, props = wptObject
		el = E(parent, "wpt", {"lat": str(lat), "lon": str(long)})
		for name in ["ele", "name"]:
			if name in props: 
				E(el, name).text = unicode(props[name])
		t = forceTime or t
		E(el, "time").text = datetime.datetime.fromtimestamp(t).isoformat()+"Z"
		for name in ["symbol", "cmt"]:
			if name in props: 
				E(el, name).text = unicode(props[name])

	def _makeTrackpoint(self, parent, tptObject, forceTime):
		t, long, lat, props = tptObject
		el = E(parent, "trkpt", {"lat": str(lat), "lon": str(long)})
		if "ele" in props:
			E(el, "ele").text = unicode(props["ele"])
		t = forceTime or t
		E(el, "time").text = datetime.datetime.fromtimestamp(t).isoformat()+"Z"
		
	def _buildElementTree(self):
		root = ElementTree.Element("gpx", {
			"xmlns":"http://www.topografix.com/GPX/1/0",
			"schemaLocation":	"http://www.topografix.com/GPX/1/0"
				" http://www.topografix.com/GPX/1/0/gpx.xsd"})
		for key in ["author", "email", "url", "urlname", "time"]:
			if self.hasProperty(key):
				E(root, key).text = self.getProperty(key)
		track = E(E(root, "trk"), "trkseg")
		for pt, forceTime in itertools.izip(self.track, self._iterTimes()):
			if "name" in pt[3]:
				self._makeWaypoint(root, pt, forceTime)
			else:
				self._makeTrackpoint(track, pt, forceTime)
		return root

	def save(self, targetName, forceTime=False):
		if not E:
			raise ReportableError("Install ElementTree to save gpx files.")
		if forceTime:
			self._setupForcedTime()
		f = open(targetName, "w")
		ElementTree.ElementTree(element=self._buildElementTree()
			).write(f, "utf-8")
		f.close()
		self.nextTime = None
	

class GPXParser(NodeBuilder):
	def _start_gpx(self, name, attrs):
		self.gpx = GPX()
	
	def _make_gpx(self, name, attrs, children):
		return self._processChildren(self.gpx, name, {
			"author": lambda val: self.gpx.setProperty("author", val),
			"email": lambda val: self.gpx.setProperty("email", val),
			"url": lambda val: self.gpx.setProperty("url", val),
			"time": lambda val: False,
			"wpt": self.gpx.addTrackpoint,
			"trk": self.gpx.addTrack,
		}, children)

	def _make_time(self, name, attrs, children):
		return time.mktime(time.strptime(
			children[0][1], "%Y-%m-%dT%H:%M:%SZ"))

	# Substitute missing times with a date in 2038.  This probably only
	# affects waypoints.  Think of something smarter until 2038 :-)
	nullTime = time.mktime((2038, 1, 1, 0, 0, 0, -1, -1, -1))

	def _make_trk(self, name, attrs, children):
		track = Track()
		return self._processChildren(track, name, {
			"name": lambda val: track.setProperty("name", val),
			"trkseg": track.addPoints,
		}, children)

	def _make_trkseg(self, name, attrs, children):
		return [c[1] for c in children]

	def _make_property(self, name, attrs, children):
		if children:
			return children[0][1]

	def _make_trkpt(self, name, attrs, children):
		time = self.nullTime
		props = {}
		for key, val in children:
			if key=="time":
				time = val
			else:
				props[key] = val
		return (time, float(attrs["lon"]), float(attrs["lat"]), props)

	_make_wpt = _make_trkpt

	_make_author = _make_email = _make_url = _make_bla = _make_symbol = \
		_make_cmt = _make_ele = _make_name = _make_property

	def _make_default(self, name, attrs, content):
		return None


def makeGPXFromXML(sourceFile):
	parser = GPXParser()
	sax.parse(sourceFile, parser)
	parser.gpx.normalize()
	return parser.gpx


if __name__=="__main__":
	g = makeGPXFromXML("foo.gpx")
	g.save("bla.gpx")
